#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX(x,y) (((x) > (y)) ? (x) : (y))

void bomb(const char *reason);
int findstr(char *puzzle, char *word, int jump, int limit);

int main(int argc, char *argv[])
{
  FILE *inp = stdin;
  char *puzzle;
  char *line;
  int i;
  int w, h;

  puts("Welcome to Andrew's wordsearch solver!");
  puts("Note that all coordinates are from (0,0)");
  puts("For backwards words, the coordinates are for the /last/ letter.\n");

  if (argc > 1)
  {
    inp = fopen(argv[1], "r");
    if(inp == NULL)
    {
      printf("Invalid filename %s.\n", argv[1]);
      exit(EXIT_FAILURE);
    }
  }
  if(fscanf(inp, "%dx%d\n\n", &w, &h) != 2) /* Eats blank line */
    bomb("Invalid file form.\n");
  puzzle = calloc(w + 1, h + 1); /* Padding used to prevent overruns */
  line   = malloc(w + 2);        /* Padding for newline and \0 */
  if(puzzle == NULL || line == NULL)
    bomb("Out of memory!");
  for(i = 0; i < h; i++)
  {
    fgets(line, w + 2, inp);
    line[w] = '\0';  /* Overwrite newline */
    strcpy(puzzle + i*(w+1), line);
  }
  free(line);
  line = malloc(MAX(w, h));
  if(line == NULL)
    bomb("Out of memory, jackass!");
  fgets(line, MAX(w,h), inp); /* Eat newline */
  while(!feof(inp))
  {
    i = findstr(puzzle, line, 1, h*(w+1));
    if(i) printf("At (%d, %d) found %s", i % (w+1), i / (w+1), line);
    i = findstr(puzzle, line, w, h*(w+1));
    if(i) printf("At (%d, %d) found %s", i % (w+1), i / (w+1), line);
    i = findstr(puzzle, line, w+1, h*(w+1));
    if(i) printf("At (%d, %d) found %s", i % (w+1), i / (w+1), line);
    i = findstr(puzzle, line, w+2, h*(w+1));
    if(i) printf("At (%d, %d) found %s", i % (w+1), i / (w+1), line);
    fgets(line, MAX(w,h), inp);
  }
  return 0;
}

void bomb(const char *reason)
{
  puts(reason);
  exit(EXIT_FAILURE);
}

int findstr(char *puzzle, char *word, int jump, int limit)
{
  int i;
  char *s = puzzle; /* Only used to compute return value */
  char *p;
  char *w = word;
  for(i = 0; i < limit; i++)
  {
    p = puzzle; /* Store starting point */
    word = w;   /* Reset word to beginning */
    while(*puzzle == *word)
    {
      puzzle += jump;
      ++word;
      if(*word == '\n')
        return p - s;
    }
    /* Search backwards */
    puzzle = p; /* Reset puzzle */
    word += strlen(word) - 2; /* Goto end-of-world */
    while(*puzzle == *word)
    {
      puzzle += jump;
      if(word == w)
        return p - s;
      --word;
    }
    puzzle = p; /* Set puzzle to starting point and move to next token */
    ++puzzle;
  }
  return 0;
}
